#title {
  display: inline;
}
/* Logo right next to the header */
#logo {
  vertical-align: middle;
  display: inline-block;
  width: 80px;
  height: 80px;
  background: url('../img/logo.png');
  background-size: contain;
  background-repeat: no-repeat;
}

#sidebar {
  max-width: 400px;
  margin: 1em;
}

#viz {
  border: 2px solid black;
  margin: 1em;
  height: 700px;
}
